# Autogenerated by Astropy's setup.py on 2018-06-01 22:33:39 UTC
from __future__ import unicode_literals
import datetime

version = "3.0.3"
githash = "1a2ca12afac85b7a0376831422ff6b1c1b5ed45e"


major = 3
minor = 0
bugfix = 3

release = True
timestamp = datetime.datetime(2018, 6, 1, 22, 33, 39)
debug = False

astropy_helpers_version = "3.0.2"

try:
    from ._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"
